// See LICENSE.SiFive for license details.

package sifive.fpgashells.devices.xilinx.xilinxml507mig

import Chisel._
import chisel3.core.{Input, Output}
import freechips.rocketchip.config.Parameters
import freechips.rocketchip.diplomacy._
import freechips.rocketchip.tilelink._
import freechips.rocketchip.util._

case class XilinxML507MIGParams(
    address: Seq[AddressSet]
)

class MemoryController extends BlackBox {
    val io = IO(new Bundle {
        val sys = new MemorySysIO
        val ddr2 = new MemoryDDR2IO

        val request_addr = Input(UInt(28.W))
        val request_read = Input(Bool())
        val request_data = Input(UInt(256.W))
        val request_mask = Input(UInt(32.W))
        val request_valid = Input(Bool())
        val request_ready = Output(Bool())

        val response_data = Output(UInt(256.W))
        val response_valid = Output(Bool())
        // no ready, as the mig does not wait
    })

    override def desiredName: String = "memory_controller"
}

class XilinxML507MIGToTL(c: XilinxML507MIGParams)(implicit p: Parameters) extends LazyModule {
    // Corresponds to MIG interface with 64 bit width and a burst length of 4
    val width = 256
    val beatBytes = width/8 // 32 byte (half a cache-line, fragmented)

    val device = new MemoryDevice
    val node = TLManagerNode(
        Seq(TLManagerPortParameters(
            Seq(TLManagerParameters(
                address         = c.address,
                resources       = device.reg,
                regionType      = RegionType.UNCACHED,
                executable      = true,
                supportsGet     = TransferSizes(1, beatBytes),
                supportsPutFull = TransferSizes(1, beatBytes),
                fifoId          = Some(0) // in-order
            )),
            beatBytes = beatBytes
        ))
    )
    // We could possibly also support supportsPutPartial, as we need support
    // for masks anyway because of the possibility of transfers smaller that
    // the data width (size signal, see below).

    lazy val module = new LazyModuleImp(this) {
        val io = IO(new Bundle {
            val port_sys = new MemorySysIO
            val port_ddr2 = new MemoryDDR2IO
        })

        val controller = Module(new MemoryController)
        io.port_sys <> controller.io.sys
        io.port_ddr2 <> controller.io.ddr2

        // in: TLBundle, edge: TLEdgeIn
        val (in, edge) = node.in(0)

        // Due to the Fragmenter defined above, all messages are 32 bytes or
        // smaller. The data signal of the TL channels is also 32 bytes, so
        // all messages will be transfered in a single beat.
        // Also, TL guarantees (see TL$4.6) that the payload of a data message
        // is always aligned to the width of the beat, e.g. in case of a 32
        // byte data signal, data[7:0] will always have address 0x***00000 and
        // data[255:247] address 0x***11111. It is also guaranteed that the
        // mask bits always correctly reflect the active bytes inside the beat
        // with respect to the size and address.
        // So we can directly forward the mask, (relative) address and possibly
        // data to the MIG interface.
        // Put requests can be acknowledged as soon as they are latched into
        // the write fifo of the MIG (possibly combinatorily).
        // For read requests, we have to store the source id and size in a
        // queue for later acknowledgment.
        // We are ready if both the MIG and the response data queue are not
        // full.

        // Widths of the A channel:
        // addressBits: 32
        // dataBits:    256
        // sourceBits:  6
        // sinkBits:    1
        // sizeBits:    3

        // source (from): in.a.bits.source
        // adresse (to): edgeIn.address(in.a.bits)
        // size: edgeIn.size(in.a.bits)
        // isPut: edgeIn.hasData(in.a.bits)

        // bits kommt von Decoupled: ready, valid + bits

        println("a parameters: " + in.a.bits.params)

        in.a.ready := Bool(false)
        in.d.valid := Bool(false)

        // Tie off unused channels
        in.b.valid := Bool(false)
        in.c.ready := Bool(true)
        in.e.ready := Bool(true)
    }
}

class XilinxML507MIG(c : XilinxML507MIGParams)(implicit p: Parameters) extends LazyModule {
    // The Fragmenter will not fragment messages <= 32 bytes, so all
    // slaves have to support this size. 64 byte specifies the maximum
    // supported transfer size that the slave side of the fragmenter supports
    // against the master (here the main memory bus). Specifying alwaysMin as
    // true results in all messages being fragmented to the minimal size
    // (32 byte). In TL1 terms, slaves correspond roughly to managers and
    // masters to clients (confusingly…).
    val fragmenter = LazyModule(new TLFragmenter(32, 64, alwaysMin=true))
    val island = LazyModule(new XilinxML507MIGToTL(c))

    val node: TLInwardNode =
        island.node := fragmenter.node

    lazy val module = new LazyModuleImp(this) {
        val io = IO(new Bundle {
            val port_sys = new MemorySysIO
            val port_ddr2 = new MemoryDDR2IO
        })

        io.port_sys <> island.module.io.port_sys
        io.port_ddr2 <> island.module.io.port_ddr2
    }
}
